﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Linq.Expressions;
using System.Reflection;
using System.Text;
using Dapper.FluentMap.Configuration;
using Fipscode.Attributes;
using Fipscode.Loader;
using Microsoft.CodeAnalysis;
using Microsoft.CodeAnalysis.CSharp;
using Microsoft.CodeAnalysis.Emit;
using DFM = Dapper.FluentMap.Mapping;

namespace Fipscode.ORM.Dapper
{
    public class MappersCreator
    {
        public Dictionary<Assembly,MetadataReference> DynamicReferences { get; } = new Dictionary<Assembly,MetadataReference>();

        public Dictionary<string,Type[]> CreateMappers(IEnumerable<Assembly> assemblies = null)
        {
            if (null == assemblies) {
                var loader = new Loader.Loader(new ProcessLoader());
                loader.Load();
                assemblies = loader.AssemblyLoaders.SelectMany(al => al.Assemblies);
            }
            var ret = new Dictionary<string, Type[]>();
            var backends = new Dictionary<string, Backend>();
            var assemblydatas = new List<AssemblyData>();
            foreach (var assembly in assemblies) {
                var assmbAttrs = assembly.GetCustomAttributes<EntityLibraryAttribute>();
                if (assmbAttrs.Count() < 1) continue;
                foreach (var assmbAttr in assmbAttrs) {
                    var backendName = assmbAttr.BackendName ?? "";
                    var convention = assmbAttr.Convention;
                    if (!backends.ContainsKey(backendName)) backends[backendName] = new Backend(backendName);
                    var backend = backends[backendName];
                    if (!backend.Assemblies.ContainsKey(assembly)) backend.Assemblies[assembly] = new AssemblyData(convention);
                }
                foreach (var type in assembly.GetTypes()) {
                    var typeAttrs = type.GetCustomAttributes<EntityAttribute>();
                    if (typeAttrs.Count() < 1) continue;
                    foreach (var typeAttr in typeAttrs) {
                        var backendName = typeAttr.BackendName ?? "";
                        var convention = typeAttr.Convention;
                        if (!backends.ContainsKey(backendName)) {
                            // no EntityLibrary with matching BackendName
                            continue;
                        }
                        var backend = backends[backendName];
                        if (!backend.Assemblies.ContainsKey(assembly)) {
                            // no EntityLibrary with matching BackendName
                            continue;
                        }
                        var assmbData = backend.Assemblies[assembly];
                        if (!assmbData.Types.ContainsKey(type)) assmbData.Types[type] = new TypeData(convention);
                    }
                    foreach (var property in type.GetProperties()) {
                        var propAttrs = property.GetCustomAttributes<FieldAttribute>();
                        var noMappings = property.GetCustomAttributes<NoMappingAttribute>();
                        var propAttrsCount = propAttrs.Count();
                        var noMappingsCount = noMappings.Count();
                        if (propAttrsCount > 0 || (propAttrsCount == 0 && noMappingsCount == 0)) {
                            if (propAttrsCount == 0) {
                                propAttrs = new[] { new FieldAttribute() };
                            }
                            foreach (var propAttr in propAttrs) {
                                var backendName = propAttr.BackendName ?? "";
                                var propName = property.Name;
                                var mapTo = propAttr.MapTo ?? propName;
                                if (!backends.ContainsKey(backendName)) {
                                    // no Entity with matching BackendName
                                    continue;
                                }
                                var backend = backends[backendName];
                                if (!backend.Assemblies.ContainsKey(assembly)) {
                                    // no Entity with matching BackendName
                                    continue;
                                }
                                var assmbData = backend.Assemblies[assembly];
                                if (!assmbData.Types.ContainsKey(type)) {
                                    // no Entity with matching BackendName
                                    continue;
                                }
                                var typeData = assmbData.Types[type];
                                if (!typeData.Fields.ContainsKey(propName)) typeData.Fields[propName] = new FieldData(mapTo);
                            }
                        }
                    }
                }
            }
            foreach (var backendKV in backends) {
                var backendName = backendKV.Key;
                var backend = backendKV.Value;
                var assembs = backend.Assemblies;
                foreach (var assembly in assembs.Keys) {
                    var types = assembs[assembly].Types;
                    var createdTypes = new List<Type>();
                    foreach (var type in types.Keys) {
                        var mappings = new Dictionary<string,string>();
                        foreach (var field in types[type].Fields.Keys) {
                            mappings[field] = types[type].Fields[field].MapTo;
                        }
                        var createdType = CreateType(type, type.Name, mappings); // TODO: pass type.Name through convention
                        if (null != createdType) {
                            createdTypes.Add(createdType);
                        }
                    }
                    ret.Add(backendName, createdTypes.ToArray());
                }
            }
            return ret;
        }

        private Type CreateType(Type type, string name, Dictionary<string,string> mappings)
        {
            var assemblyName = type.Assembly.FullName;
            var nspace = type.Namespace;
            var sb = new StringBuilder();
            foreach (var kv in mappings) {
                sb.Append($"            Map(m => m.{kv.Key}).ToColumn(\"{kv.Value}\");\n");
            }
            var code = $@"using Dapper.FluentMap.Mapping;
using {nspace};

namespace Fipscode.ORM.Dapper.{nspace}
{{
    public class {name}Mapper : EntityMap<{name}>
    {{
        public {name}Mapper()
        {{
{sb.ToString()}        }}
    }}
}}";
            var assembly = Compile(code, assemblyName, new MetadataReference[] {
                MetadataReference.CreateFromFile(type.Assembly.Location),
                MetadataReference.CreateFromFile(typeof(DFM.EntityMap<Type>).Assembly.Location),
            });
            Type ret = assembly.GetType($"Fipscode.ORM.Dapper.{nspace}.{name}Mapper");
            return ret;
        }

        public Action<FluentMapConfiguration> CreateInitializer(Type[] types)
        {
            var assemblyName = "TODO";
            var l = "Fipscode.ORM.Dapper.".Length;
            var sb = new StringBuilder();
            foreach (var type in types) {
                sb.Append($"            c.AddMap(new {type.FullName.Substring(l)}());\n");
            }
            var code = $@"using Dapper.FluentMap.Configuration;

namespace Fipscode.ORM.Dapper
{{
    public class FluentMapperInitializer : FluentMapperInitializerBase
    {{
        public override void Initialize(FluentMapConfiguration c)
        {{
{sb.ToString()}        }}
    }}
}}";
            var assemblies = types.Select(type => type.Assembly).ToArray();
            var references = new List<MetadataReference>(assemblies
                .Select(asm => DynamicReferences[asm])
            );
            references.AddRange(new[]{
                MetadataReference.CreateFromFile(typeof(DFM.EntityMap<Type>).Assembly.Location),
                MetadataReference.CreateFromFile(typeof(FluentMapperInitializerBase).Assembly.Location),
            });
            var assembly = Compile(code, assemblyName, references.ToArray());
            Type ret = assembly.GetType("Fipscode.ORM.Dapper.FluentMapperInitializer");
            var inst = Activator.CreateInstance(ret);
            return ((FluentMapperInitializerBase)inst).Initialize;
        }

        private Assembly Compile(string code, string assemblyName, MetadataReference[] references)
        {
            Console.WriteLine(code);
            var defaultReferences = new MetadataReference[] {
                MetadataReference.CreateFromFile(typeof(Expression<Type>).Assembly.Location),
                MetadataReference.CreateFromFile(typeof(object).Assembly.Location),
                MetadataReference.CreateFromFile(Assembly.Load("netstandard").Location),
                MetadataReference.CreateFromFile(Assembly.Load("System.Runtime").Location),
            };
            var passReferences = new MetadataReference[defaultReferences.Length + references.Length];
            defaultReferences.CopyTo(passReferences, 0);
            references.CopyTo(passReferences, defaultReferences.Length);
            SyntaxTree syntaxTree = CSharpSyntaxTree.ParseText(code);

            CSharpCompilation compilation = CSharpCompilation.Create(
                assemblyName,
                new[] { syntaxTree },
                passReferences,
                new CSharpCompilationOptions(OutputKind.DynamicallyLinkedLibrary)
            );
            using (var memoryStream = new MemoryStream()) {
                EmitResult result = compilation.Emit(memoryStream);
                if (result.Success) {
                    memoryStream.Seek(0, SeekOrigin.Begin);
                    Assembly assembly = Assembly.Load(memoryStream.ToArray());
                    DynamicReferences.Add(assembly, MetadataReference.CreateFromStream(memoryStream));
                    return assembly;
                } else {
                    foreach (var diagnostic in result.Diagnostics) {
                        Console.WriteLine(diagnostic);
                    }
                    return null;
                }
            }
        }
    }

    public class FluentMapperInitializerBase
    {
        public virtual void Initialize(FluentMapConfiguration c) {}
    }
}
